import typer

import slh_sh.commands.extract as extract
import slh_sh.commands.get as get
import slh_sh.commands.add as add
import slh_sh.commands.go as go
import slh_sh.commands.sync as sync
import slh_sh.commands.self as self

from slh_sh.commands.query import query

app = typer.Typer()

app.add_typer(extract.app, name="extract")
app.add_typer(get.app, name="get")
app.add_typer(add.app, name="add")
app.add_typer(go.app, name="go")
app.add_typer(sync.app, name="sync")
app.add_typer(self.app, name="self")

app.command()(query)

if __name__ == "__main__":
    app()



import fitz
import math

from itertools import groupby


def get_pdf_text(page, rect):
    """Return text containted in the given rectangular highlighted area.

    Args:
        page (fitz.page): the associated page.
        rect (fitz.Rect): rectangular highlighted area.
    """
    words = page.get_text("blocks")  # list of words on page
    words.sort(key=lambda w: (w[3], w[0]))  # ascending y, then x
    mywords = [w for w in words if fitz.Rect(w[:4]).intersects(rect)]
    group = groupby(mywords, key=lambda w: w[3])
    for y1, gwords in group:
        pdf_text = " ".join(w[4] for w in gwords)
        pdf_text = pdf_text.replace("\n", " ").strip()
        return pdf_text


def rgb_to_hex(rgb):
    """Converts an RGB tuple to a hex string.

    Args:
    rgb: A tuple of 3 floats, representing the red, green, and blue components
        of the color, in the range [0, 1].

    Returns:
    A string representing the color in hexadecimal format.
    """

    r, g, b = rgb
    hex_r = int(r * 255)
    hex_g = int(g * 255)
    hex_b = int(b * 255)
    return f"#{hex_r:02x}{hex_g:02x}{hex_b:02x}"


def hex_to_rgb(hex_color):
    """
    Converts a hex color to RGB.

    Args:
        hex_color: A string representing the hex code of the color.

    Returns:
        A tuple of three integers representing the RGB values of the color.
    """

    # Split the hex color into three parts.
    red, green, blue = hex_color[1:3], hex_color[3:5], hex_color[5:7]

    # Convert each part from hexadecimal to decimal.
    red = int(red, 16)
    green = int(green, 16)
    blue = int(blue, 16)

    # Return the RGB values as a tuple.
    return (red, green, blue)


def is_color_close(rgb_color, hex_color, threshold=50):
    """
    Checks if a RGB color is close enough to the provided hex color.

    Args:
        rgb_color: A tuple of three integers representing the RGB values of the color.
        hex_color: A string representing the hex code of the color.
        threshold: The maximum distance between the two colors for them to be considered close.

    Returns:
        True if the two colors are close enough, False otherwise.
    """

    rgb_of_config_hex = hex_to_rgb(hex_color)
    rgb_fixed = tuple(int(color * 255) for color in rgb_color)

    # Calculate the distance between the two colors.
    distance = math.sqrt(
        (rgb_fixed[0] - rgb_of_config_hex[0]) ** 2
        + (rgb_fixed[1] - rgb_of_config_hex[1]) ** 2
        + (rgb_fixed[2] - rgb_of_config_hex[2]) ** 2
    )

    # Return True if the distance is less than or equal to the threshold.
    return distance <= threshold



import typer
import os

from rich import print
from pathlib import Path
from typing_extensions import Annotated

from slh_sh.utils.file import get_pdf_dir, get_file_path, get_conf
from slh_sh.utils.log import logger, get_now
from slh_sh.modules.extract import (
    extract_cit,
    extract_bib,
    extract_dl,
    extract_filename,
    extract_keywords,
    extract_annots,
    extract_dist,
    extract_total_dist_sheet_sync,
    extract_dist_ws_sheet_sync,
)

app = typer.Typer()

##
## Citation
##


@app.command("cit")
def cit(
    db: Annotated[bool, typer.Option(help="Save to SQLite database file")] = False,
):
    """Extracts the citations from the database"""
    input(
        f"""
        Press Enter to generate citations:

        Format: APA 7
        Google Sheet's URL: {get_conf("gs_url")}

        Press Ctrl+C to cancel.
        """
    )

    print(f"Extracting citations from db...")

    citations, authorNoneRemoved = extract_cit()

    print(citations)
    print(f"{len(citations)} citations added to the database")
    print(authorNoneRemoved)


##
## Bibliography
##


@app.command()
def bib(
    csv: Annotated[str, typer.Argument(help="Covidence CSV Export")] = get_conf(
        "csv_export"
    ),
    db: Annotated[bool, typer.Option(help="Save to SQLite database file")] = False,
):
    """Extracts the bibliographies from the CSV export"""
    input(
        f"""
        Press Enter to generate bibliographies:

        CSV File: {csv}
        Format: APA 7
        Google Sheet's URL: {get_conf("gs_url")}

        Press Ctrl+C to cancel.
        """
    )

    bibs = extract_bib(csv, db)

    print("Bibliographies added to the database:")
    print(bibs)


##
## Download
##


@app.command()
def dl(
    html: Annotated[
        str,
        typer.Argument(
            help="HTML export containing Covidence Number and Download Links"
        ),
    ] = get_conf("html_export"),
    pdfdir: Annotated[str, typer.Argument(help="Directory to save PDFs")] = get_conf(
        "pdf_path"
    ),
    html_id_element: Annotated[
        str,
        typer.Argument(
            help="Class name of the 'div' element containing Covidence Number or ID in a div"
        ),
    ] = get_conf("html_id_element"),
    html_dl_class: Annotated[
        str,
        typer.Argument(
            help="Class name of the 'a' element containing the URL of the PDF"
        ),
    ] = get_conf("html_dl_class"),
):
    """Downloads the PDFs from the HTML export"""
    input(
        f"""
        Press Enter to download PDFs:

        HTML export File: {html}
        Study Folder: {pdfdir}

        Press Ctrl+C to cancel.
        """
    )
    print(
        f"""
            Downloading PDFs from {html}...
            """
    )
    # Create the PDF directory if it doesn't exist.
    pdf_dir = Path.cwd() / pdfdir
    if not pdf_dir.is_dir():
        pdf_dir.mkdir()
    study_headers = extract_dl(html, pdf_dir, html_id_element, html_dl_class)
    print(
        f"""
        :tada: {len(study_headers)} PDFs from {html} to {pdf_dir} downloaded.
        """
    )


##
## Filename
##


@app.command()
def filename(
    csv: Annotated[str, typer.Argument(help="Covidence CSV Export")] = get_conf(
        "csv_export"
    ),
    rename: Annotated[bool, typer.Option(help="Also Rename the PDFs")] = False,
    db: Annotated[bool, typer.Option(help="Save to SQLite database file")] = False,
):
    """Extracts the filenames from the PDFs"""

    print(f"Extracting filenames from {csv}...")
    fileNames = extract_filename(csv, rename, db)
    print(fileNames)
    print(
        f"Updated database with {len(fileNames)} filenames on studies Table, Filenames column..."
    )
    if rename:
        print(
            f"Renamed {len(fileNames)} PDFs in {get_conf('pdf_path')} folder with the filenames..."
        )


##
## Keywords
##


@app.command()
def keywords(
    id: Annotated[str, typer.Option(help="ID e.g. Covidence number to extract keywords")] = "",
    all: Annotated[
        bool,
        typer.Option(help="Extract keywords from all PDFs in pdf_path folder"),
    ] = False,
    db: Annotated[bool, typer.Option(help="SQLite database file")] = False,
):
    """Extracts the keywords from the PDFs"""
    print(f"Keywords {id}...")

    pdf_dir = get_pdf_dir()
    pdf_path = None

    if all:
        for file_name in os.listdir(pdf_dir):
            id = file_name.split("_")[0].remove("#")
            pdf_path = get_file_path(id)
            all_keywords = extract_keywords(id, pdf_path, db)
        print(all_keywords)
        print(
            f"Extracted keywords from {len(all_keywords)} PDFs and added them to the Database..."
        )
    elif id != "":
        pdf_path = get_file_path(id)
        keywords = extract_keywords(id, pdf_path, db)
        print(
            f"Extracted keywords {keywords} from {pdf_path} and added them to the Database..."
        )
    else:
        print(
            "Please enter an ID, e.g. covidence number using --id [Covidence Number] or use --all"
        )


##
## Annotations
##


@app.command()
def annots(
    id: Annotated[str, typer.Option(help="ID, e.g. Covidence number to extract keywords")] = "",
    color: Annotated[
        str, typer.Option(help="Color of the annotations to extract")
    ] = "",
    all: Annotated[
        bool, typer.Option(help="Extract annotations from all PDFs in pdf_path folder")
    ] = False,
    db: Annotated[bool, typer.Option(help="Save to SQLite database file")] = False,
):
    """Extracts the annotations from the PDFs"""
    print(f"Fetching Annotations of {color} (themes,topic,colored texts) from {id}...")

    pdf_dir = get_pdf_dir()
    pdf_path = None

    if all:
        for file_name in os.listdir(pdf_dir):
            id = file_name.split("_")[0].remove("#")
            pdf_path = get_file_path(id)
            res = extract_annots(id, color, pdf_path, db)
            print(res)
    elif id != "":
        pdf_path = get_file_path(id)
        res = extract_annots(id, color, pdf_path, db)
        print(res)
    else:
        print(
            """
[bold red]Please enter an ID, e.g. Covidence Number using --id [Covidence Number] or use --all[/bold red]
            """
        )


##
## Distribution
##


@app.command()
def dist(
    term: Annotated[str, typer.Argument(help="Search term")] = "",
    id: Annotated[str, typer.Option(help="ID, e.g. Covidence number")] = "",
    output: Annotated[
        str,
        typer.Option(help="Output format, available options are Json, YAML and CSV"),
    ] = "",
    db: Annotated[bool, typer.Option(help="Save to SQLite database file")] = False,
    all: Annotated[bool, typer.Option(help="All PDFs")] = False,
    tdsheet: Annotated[
        bool,
        typer.Option(
            help="Apply to Distribution Column on default_studies Worksheet of Google Sheet"
        ),
    ] = False,
    wsdsheet: Annotated[
        bool, typer.Option(help="Apply to Distribution Worksheet on Google Sheet")
    ] = False,
):
    """Extracts the distribution of a search term in the PDFs"""

    if tdsheet == True and all == True and term == "" and id == "":
        res_total_dist_col = extract_total_dist_sheet_sync()
        print(
            f"Total distribution column update on Google Sheet Studies worksheet from db, {res_total_dist_col}"
        )
    elif wsdsheet == True and all == True and term == "" and id == "":
        res_dist_ws = extract_dist_ws_sheet_sync()
        print(
            f"Total distribution worksheet update on Google Sheet from db, {res_dist_ws}"
        )
        # res_dist_ws = extract_dist_ws_sheet_sync()
    elif all and id == "" and term != "":
        pdf_dir: str = get_pdf_dir()
        for file_name in os.listdir(pdf_dir):
            id = file_name.split("_")[0].replace("#", "")
            pdf_path: str = get_file_path(id)
            if pdf_path.endswith(".pdf"):
                total_count, dist_list = extract_dist(pdf_path, term, id, db)
                msg = {
                    "total_count": total_count,
                    "dist_list": dist_list,
                }
                print(msg)
            else:
                logger().info(f"{get_now()} {pdf_path} is not a PDF file.")
    elif id != "" and term != "" and all == False:
        pdf_path = get_file_path(id)
        total_count, dist_list = extract_dist(pdf_path, term, id, db)
        msg = {
            "total_count": total_count,
            "dist_list": dist_list,
        }
        print(msg)
    elif id != "" and term == "" and all == False and wsdsheet == True:
        res_dist_ws = extract_dist_ws_sheet_sync(id)
        print(
            f"Total distribution worksheet update on Google Sheet from db, {res_dist_ws}"
        )
    else:
        print(
            """
[bold red]Please enter an ID, e.g. covidence number using --id [Covidence Number] or use --all[/bold red]
            """
        )



from __future__ import annotations
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy import ForeignKey


from slh_sh.utils.db import BaseModel


class Study(BaseModel):
    __tablename__ = "studies"

    id: Mapped[int] = mapped_column(
        index=True, primary_key=True, autoincrement=True, nullable=False
    )
    created_at: BaseModel.created_at
    updated_at: BaseModel.updated_at
    title: Mapped[str] = mapped_column(index=True, nullable=False)
    authors: Mapped[str] = mapped_column(index=True, nullable=False)
    abstract: Mapped[str] = mapped_column(index=True, nullable=False)
    published_year: Mapped[int] = mapped_column(index=True, nullable=False)
    published_month: Mapped[int] = mapped_column(index=True, nullable=True)
    journal: Mapped[str] = mapped_column(index=True, nullable=True)
    volume: Mapped[str] = mapped_column(index=True, nullable=True)
    issue: Mapped[str] = mapped_column(index=True, nullable=True)
    pages: Mapped[str] = mapped_column(index=True, nullable=True)
    accession_number: Mapped[str] = mapped_column(index=True, nullable=True)
    doi: Mapped[str] = mapped_column(index=True, nullable=True)
    ref: Mapped[str] = mapped_column(index=True, nullable=True)
    Covidence: Mapped[int] = mapped_column(index=True, nullable=True)
    study: Mapped[str] = mapped_column(index=True, nullable=True)
    notes: Mapped[str] = mapped_column(index=True, nullable=True)
    tags: Mapped[str] = mapped_column(index=True, nullable=True)
    filename: Mapped[str] = mapped_column(index=True, nullable=True)
    keywords: Mapped[str] = mapped_column(index=True, nullable=True)
    citation: Mapped[str] = mapped_column(index=True, nullable=True)
    bibliography: Mapped[str] = mapped_column(index=True, nullable=True)
    full_text: Mapped[str] = mapped_column(index=True, nullable=True)
    total_annotations: Mapped[int] = mapped_column(index=True, nullable=True)
    total_distribution: Mapped[int] = mapped_column(index=True, nullable=True)


class Theme(BaseModel):
    __tablename__ = "themes"

    id: Mapped[int] = mapped_column(
        index=True, primary_key=True, autoincrement=True, nullable=False
    )
    created_at: BaseModel.created_at
    updated_at: BaseModel.updated_at
    color: Mapped[str] = mapped_column(index=True, nullable=False)
    term: Mapped[str] = mapped_column(index=True, nullable=False)
    hex: Mapped[str] = mapped_column(index=True, nullable=False)


class Annotation(BaseModel):
    __tablename__ = "annotations"

    id: Mapped[int] = mapped_column(
        index=True, primary_key=True, autoincrement=True, nullable=False
    )
    created_at: BaseModel.created_at
    updated_at: BaseModel.updated_at
    studies_id: Mapped[int] = mapped_column("Study", ForeignKey("studies.id"))
    theme_id: Mapped[int] = mapped_column("Theme", ForeignKey("themes.id"))
    count: Mapped[int] = mapped_column(index=True, nullable=False)
    page_number: Mapped[int] = mapped_column(index=True, nullable=False)
    annot_rgb_color: Mapped[str] = mapped_column(index=True, nullable=False)
    annot_hex_color: Mapped[str] = mapped_column(index=True, nullable=False)
    text: Mapped[str] = mapped_column(index=True, nullable=False)


class Distribution(BaseModel):
    __tablename__ = "distribution"

    id: Mapped[int] = mapped_column(
        index=True, primary_key=True, autoincrement=True, nullable=False
    )
    created_at: BaseModel.created_at
    updated_at: BaseModel.updated_at
    studies_id: Mapped[int] = mapped_column("studies_id", ForeignKey("studies.id"))
    theme_id: Mapped[int] = mapped_column("theme_id", ForeignKey("themes.id"))
    count: Mapped[int] = mapped_column(index=True, nullable=False)
    page_number: Mapped[int] = mapped_column(index=True, nullable=False)
    term: Mapped[str] = mapped_column(index=True, nullable=False)
    text: Mapped[str] = mapped_column(index=True, nullable=False)



import typer
import os
import yaml


def saveConfigFile(
    project_name, gs_url, csv_export, html_export, google_credentials, sqlite_db
):
    """Generates and saves config.yaml file

    Returns:
        str: message
    """
    config = {
        "project_name": project_name,
        # URL to Google Sheet
        "gs_url": gs_url,
        # URL to Google Drive
        "gd_url": "",
        # Name of the default CSV file
        "csv_export": csv_export,
        # Name of the default HTML file
        "html_export": html_export,
        # Name of id of html element
        "html_id_element": "study-header",
        # Name of download class in html element
        "html_dl_class": "action-link download",
        # Name of the default Google Credentials file
        "google_credentials": google_credentials,
        # Name of the default SQLite Database file
        "sqlite_db": sqlite_db,
        # Name of the default PDF folder
        "pdf_path": "studies_pdf",
        # Number of the header row in Google Sheet
        "gs_header_row_number": "3",
        # Name of the Studies sheet in Google Sheet and name of the default table in SQLite Database
        "default_studies": "Studies",
        # Name of the default ID column in Google Sheet and SQLite Database
        "default_id": "Covidence",
        # Themes, Searches, and Sources will be added to database with sync yaml command
        # Themes (colors) for the annotations
        "themes": {
            # Theme name (color)
            "blue": {
                # Theme color in hex
                "hex": "#3389FF",
                # Theme annotation (Topic)
                "term": "Challenges in AI Laws",
            },
            "red": {
                "hex": "#FF3333",
                "term": "Challenges in AI Ethics",
            },
            "vio": {
                "hex": "#CC33FF",
                "term": "Challenges in AI Politics",
            },
        },
        # Searches (keywords) for the studies
        "searches": {
            "search_1": "AI and Regulations",
            "search_2": "AI, Regulations",
        },
        # Sources (where the study is found)
        "sources": {
            "source_1": "Google Scholar",
            "source_2": "Scopus",
            "source_3": "University Library",
        },
    }

    config_yaml_path = os.path.join(".", "config.yaml")
    if os.path.exists(config_yaml_path):
        yes = typer.prompt(
            f"""

            Config file already exists at {config_yaml_path}.

            Enter (y) to overwrite the file.

            Press Ctrl+C to exit.

            """
        )
        if yes == "y":
            with open(config_yaml_path, "w") as f:
                f.write(
                    """#
# Config file for slh
#
# project_name - Name of the project directory
# gs_url - URL to Google Sheet
# gd_url - URL to Google Drive
# csv_export - Name of the default CSV file
# html_export - Name of the default HTML file
# html_id_element - Name of id of html element
# html_dl_class - Name of download class in html element
# google_credentials - Name of the default Google Credentials file
# sqlite_db - Name of the default SQLite Database file
# pdf_path - Full path of the PDF folder e.g /Users/Sara/Downloads/uni/Thesis/SLR_Thesis_Lib
# gs_header_row_number - Number of the header row in Google Sheet
# default_studies - Name of the Studies sheet in Google Sheet and name of the default table in SQLite Database
# default_id - Name of the default ID column in Google Sheet and SQLite Database
# themes - Themes (colors) for the annotations
# searches - Searches (keywords) for the studies
# sources - Sources (where the study is found)
#
# Themes, Searches, and Sources will be added to database with sync yaml command
#
# yamllint disable rule:line-length
"""
                )
                yaml.dump(
                    config,
                    f,
                    indent=4,
                    width=100,
                    allow_unicode=True,
                    default_flow_style=False,
                    explicit_start=True,
                    encoding="utf-8",
                    sort_keys=False,
                    line_break=None,
                )
            return f"Config file created at {config_yaml_path}"



import typer
import time

from pathlib import Path
from rich import print

from slh_sh.utils.config import saveConfigFile

# from slh_sh.utils.db import create_db

slh_version: str = "0.1.9"

app = typer.Typer()

default_project_name: str = (
    f"slr-ai-thesis-{time.strftime('%Y')}-{time.strftime('%m')}-{time.strftime('%d')}"
)


@app.command()
def version():
    print(
        f"""

            SLRs Little Helper (slh) {slh_version}

        """
    )


@app.command()
def list():
    """List all available commands and their descriptions."""
    print(
        f"""

            SLRs Little Helper (slh) {slh_version}

            Usage: slh-sh [OPTIONS] COMMAND [ARGS]...

            Example:
                # slh-sh --help
                # slh-sh self --help
                # slh-sh extract cit --help

            Commands:
                - self
                    - init      # Initializes the project, questionaire or default config file
                    - list      # Lists all available commands
                    - check     # Checks if there is a new version
                    - update    # Updates to the latest version.
                    - backup    # Creates compressed backup file from the project directory.
                    - restore   # Restores the project directory from the backup file
                    - logs      # Shows the logs
                    - version   # Shows the version
                - add
                    - csv       # Loads studies.csv into sqlite database.
                - extract
                    - cit       # Extracts and generates APA 7 citation from the file name in db and updates the citation column in the studies table.
                    - bib       # Extracts the bibliography from the csv file and updates the bibliography column in the studies table.
                    - dl        # Extracts the download link from the html file and downloads the pdf files.
                    - filename  # Extracts the filename from the csv file and updates the filename column in the studies table.
                    - keywords  # Extracts the keywords from the pdf file and updates the keywords column in the studies table.
                    - annots    # Extracts the annotations from the pdf file and updates the annotations table in the database.
                    - dist      # Extracts the distribution of the term from the pdf file and updates the distribution table in the database.
                - sync
                    - update    # Update the Google Sheets with data from Database.
                    - fetch     # Fetch and save data from Google Sheet to a new database table.
                    - config    # Iterate over Themes, Searches, and Sources in config.yaml and insert into database.
                - get
                    - info      # Get info about a study and its Citation and Bibliography from a database table by ID.
                - query         # Get themes about a study from a database table by ID.
                - go
                    - gd        # Opens the Google Drive folder.
                    - gs        # Opens the Google Sheet.
                    - pdf       # Opens a PDF file in the default PDF reader.
                    - db        # Opens the SQLite database file in the default database viewer.
                    - doi       # Opens the DOI URL in the default browser.
        """
    )


@app.command()
def check():
    print("Check... - Not Implemented Yet")


@app.command()
def update():
    print("Update... - Not Implemented Yet")


@app.command()
def backup():
    print("Backup... - Not Implemented Yet")


@app.command()
def restore():
    print("Restore... - Not Implemented Yet")


@app.command()
def logs():
    print("Logs... - Not Implemented Yet")


@app.command()
def init(config: bool = typer.Option(False, help="Only create config file")):
    print(
        """

        Good day researcher! :wave:

        Welcome to slh-sh :space_invader:

        It stands for SLRs Little Helper

        A CLI tool to help you with your Systematic Literature Review. :books:

        Visit https://slh.sh for more info.

        """
    )
    if config:
        res = saveConfigFile(
            default_project_name,
            "",
            "studies.csv",
            "export.html",
            "credentials.json",
            "slh.db",
        )
        print(res)
        return
    else:
        if typer.confirm(
            """

            Choose (Y) if you want to start initialization questionaire.

            Choose (n) to initialize with the default config file.

            """
        ):
            project_name = typer.prompt(
                """

                What is the name of your project? (e.g. slr-2023-4-19)

                    [ ] - A folder with this name will be created in your current directory.

                    Press Enter to use the default name.

                """,
                default=f"slr-ai-thesis-{time.strftime('%Y')}-{time.strftime('%m')}-{time.strftime('%d')}",
            )

            csv_export = typer.prompt(
                """

                What is the CSV filename?

                    e.g. studies.csv exported from Covidence

                    Press Enter to use the default name.

                """,
                default="studies.csv",
            )

            gs_url = typer.prompt(
                """

                What is the Google Sheet's URL?

                """
            )

            html_export = typer.prompt(
                """

                What is the HTML Export file name?

                    e.g. export.html containing ID and Download links

                    Press Enter to use the default name.

                """,
                default="export.html",
            )

            google_credentials = typer.prompt(
                """

                What is the name to your Google Credentials?

                    e.g. credentials.json

                    Press Enter to use the default name.

                """,
                default="credentials.json",
            )

            sqlite_db = typer.prompt(
                """

                What is the name to your SQLite Database?

                    e.g. slh.db

                    [ ] - It will be created by init command.

                    Press Enter to use the default name.

                """,
                default="slh.db",
            )

            # Create the project directory
            project_dir = Path.cwd() / project_name
            if not project_dir.is_dir():
                project_dir.mkdir()

            # # Create the sqlite database
            # sqlite_db_path = project_dir / sqlite_db
            # create_db(sqlite_db_path)

            if typer.confirm(
                """

                Do you want to create a new config file?

                """
            ):
                saveConfigFile(
                    project_name,
                    gs_url,
                    csv_export,
                    html_export,
                    google_credentials,
                    sqlite_db,
                )
        else:
            res = saveConfigFile(
                default_project_name,
                "",
                "studies.csv",
                "export.html",
                "credentials.json",
                "slh.db",
            )
            print(res)

        print(
            f"""

            :star: Project directory created at {project_dir}
            :star: Config file created at {project_dir}/config.yaml
            :star: sqlite database created at {project_dir}/{sqlite_db}

            [bold]Hooray! :space_invader:[/bold]

            Next Steps:

                - [yellow]Move credentials.json to the {project_dir}[/yellow]
                - [yellow]Move studies.csv to the {project_dir}[/yellow]
                - [yellow]Move export.html to the {project_dir}[/yellow]
                # [yellow]cd {project_dir}[/yellow]
                # [yellow]slh --help[/yellow]
                # [yellow]slh load csv[/yellow] # to load studies.csv into sqlite database

            """
        )



import os
import time
import re
import fitz
import requests
import pandas as pd

import sys

from bs4 import BeautifulSoup
from pdfminer.high_level import extract_text

from slh_sh.utils.log import logger
from slh_sh.utils.db import get_db
from slh_sh.utils.file import file_name_generator, get_conf, get_random_string
from slh_sh.utils.pdf import (
    rgb_to_hex,
    get_pdf_text,
    is_color_close,
)
from slh_sh.modules.sync import (
    update_sheet_cell,
    get_spreadsheet_by_url,
    get_worksheet_by_name,
    get_worksheet_id_col_index_values,
    get_worksheet_updating_col_index_header,
    get_worksheet_headers_row_values,
    create_new_worksheet,
)
from slh_sh.data.models import (
    Study,
    Theme,
    Annotation,
    Distribution,
)


##
## Extract Citation
##


def extract_cit(db=False):
    """Extracts and generates APA 7 citation from the file name in db and updates the citation column in the studies table

    Returns:
        list, list: List of citations, List of rows with None values
    """

    dbs = get_db()
    rows = dbs.query(Study).all()

    citations = []
    authorNoneRemoved = []
    for row in rows:
        if None in row:
            print(f"{row} is none")
            authorNoneRemoved.append(row)
        else:
            fileName: str = file_name_generator(row[0], row[1], row[2])

            fileName = fileName.split("_")

            year = fileName[-1]
            if len(fileName) == 3:  # 1 author
                authorsName = fileName[1]
                citation = f"({authorsName}, {year})"
            elif len(fileName) == 4:  # 2 authors
                authorsName = f"{fileName[1]} & {fileName[2]}"
                citation = f"({authorsName}, {year})"
            elif len(fileName) == 5:  # more than 2 authors using first author and et al
                authorsName = f"{fileName[1]} et al."
                citation = f"({authorsName}, {year})"

            citation = f"({authorsName}, {row[2]})"

            if db:
                dbs.query(Study).filter(Study.Covidence == fileName[0]).update(
                    {Study.citation: citation}
                )
                dbs.commit()
            citations.append(citation)

    return citations, authorNoneRemoved


##
## Extract Bibliography
##


def extract_bib(csv, db=False):
    """Extracts the bibliography from the csv file and updates the bibliography column in the studies table

    Args:
        csv (file): csv file

    Returns:
        list: List of bibliographies
    """
    csv_df = pd.read_csv(csv)
    csv_df["Published Year"] = csv_df["Published Year"].astype(str)
    csv_df = csv_df.fillna("None")

    dbs = get_db()

    bibs = []

    db_res = dbs.query(Study).all()

    for row in db_res:
        bib = {
            "covidence_number": row.Covidence,
            "author": row.authors,
            "year": row.published_year,
            "title": row.title,
            "journal": row.journal,
            "volume": row.volume,
            "issue": row.issue,
            "pages": row.pages,
        }

        bibs.append(bib)

    for bib in bibs:
        bibliography: str = f"{bib['author']} ({bib['year']}). {bib['title']}. {bib['journal']} {bib['volume']}({bib['issue']}), {bib['pages']}."
        id: str = f"{bib['covidence_number']}"

        if db:
            dbs.query(Study).filter(Study.Covidence == id).update(
                {Study.bibliography: bibliography}
            )
            dbs.commit()

    dbs.close()

    return bibs


##
## Extract Download
##


def extract_dl(html, pdf_dir, html_id_element, html_dl_class):
    """Extracts the download link from the html file and downloads the pdf files

    Args:
        html (file): html file name from config.yaml
        pdf_dir (str): pdf folder name from config.yaml
        html_id_element (str): id element of html_dl_class name from cli option or config.yaml
        html_dl_class (str): download link element that contains a element and a URL, from cli option or config.yaml

    Raises:
        Exception: Bad response status code, remove the section of the HTML file containing this link and try again.

    Returns:    TODO: return the pdf file name
        Soup result: result of the soup.find_all
    """
    with open(html, "r") as f:
        html_string = f.read()

    soup = BeautifulSoup(html_string, "html.parser")

    study_headers = soup.find_all("div", class_=html_id_element)

    for study_header in study_headers:
        study_header_number = re.findall("#(\d+)", study_header.text)[0]
        download_link_element = study_header.parent.find("a", class_=html_dl_class)
        download_link = download_link_element["href"]

        print(
            f"""
            :runner: Downloading PDF with ID {study_header_number} and URL {download_link}..."""
        )

        pdf_path = os.path.join(pdf_dir, f"{study_header_number}.pdf")

        if not os.path.exists(pdf_path):
            response = requests.get(download_link)
            if response.status_code == 200:
                with open(pdf_path, "wb") as f:
                    f.write(response.content)
                print(f"PDF downloaded: {pdf_path}")
                time.sleep(3)
            else:
                raise Exception(
                    f"""
                    [bold red]Error[/bold red]: {response.status_code}

                    Failed to download PDF: {download_link}, remove the section of the HTML file containing this link and try again.
                    """
                )
        else:
            print(f"PDF already exists: {pdf_path}")

    return study_headers


##
## Extract Filename
##


def extract_filename(csv, rename=False, db=False):
    """Extracts the filename from the csv file and updates the filename column in the studies table

    Args: TODO: add name convention options
        csv (file): csv file
        rename (bool): rename the pdf file

    Returns:
        list: List of file names
    """
    csv_df = pd.read_csv(csv)
    csv_df["Published Year"] = csv_df["Published Year"].astype(str)
    csv_df = csv_df.fillna("None")

    file_names = []
    for i in csv_df["Authors"]:
        # select covidence # from csv_df where authors is i
        covidence_number: str = csv_df.loc[csv_df["Authors"] == i][
            "Covidence #"
        ].values[0]
        # select publicaiton year from csv_df where authors is i
        year: str = csv_df.loc[csv_df["Authors"] == i]["Published Year"].values[0]

        file_name: str = file_name_generator(covidence_number, i, year)

        if db:
            dbs = get_db()
            dbs.query(Study).filter(Study.Covidence == covidence_number).update(
                {Study.filename: file_name}
            )
            dbs.commit()
            dbs.close()

        if rename:
            pdf_path = os.path.join(get_conf("pdf_path"), f"{covidence_number}.pdf")
            if os.path.exists(pdf_path) and not pdf_path.endswith(f"{file_name}.pdf"):
                os.rename(
                    pdf_path, os.path.join(get_conf("pdf_path"), f"{file_name}.pdf")
                )
            else:
                print("file exists")

        file_names.append(file_name)

    return file_names


##
## Extract Keywords
##


def extract_keywords(id, pdf_path, db=False):
    """Extracts the keywords from the pdf file and updates the keywords column in the studies table

    Args:
        id (int): ID, e.g. Covidence number
        pdf_path (path): Path to the pdf file
        db (bool, optional): Save to database? Defaults to False.

    Returns:
        _type_: _description_
    """
    print(f"Extracting keywords of: {pdf_path}...")

    text = extract_text(pdf_path)
    regex = r"(?is)(keyword.*?)(?:(?:\r*\n){2})"
    matches = re.findall(regex, text)
    length = len(matches)

    all_keywords = []
    if length == 0:
        print(f"Keywords not found for {pdf_path}")
        all_keywords.append(f"{id} None")
    elif length >= 1:
        keywords = (
            matches[0]
            .replace("\n", " ")
            .replace("Keywords:", "")
            .replace("KEYWORDS:", "")
            .replace("Keywords", "")
            .replace("KEYWORDS", "")
            .strip()
        )
        all_keywords.append(f"{id} {keywords}")
        print(id, keywords)
        if db:
            dbs = get_db()
            dbs.query(Study).filter(Study.Covidence == id).update(
                {Study.keywords: keywords}
            )
            dbs.commit()
            dbs.close()

    return all_keywords


##
## Extract Annotations
##


def extract_annots(id: int, color: str, pdf_path: str, db: bool = False):
    """Extracts the annotations from the pdf file and updates the annotations table in the database

    Args:
        id (int): ID, e.g. Covidence number
        color (str): Color of the annotations to extract
        pdf_path (str): Path to the pdf file
        db (bool, optional): Defaults to False

    Returns:
        total_count (int), return_list (list): Total count of the annotations found in the pdf file, and the list of the annotations
    """
    total_count = 0
    page_annots = []
    page_number = 0
    return_list = []

    with fitz.open(pdf_path) as doc:
        for page in doc:
            count = 0
            page_number = page.number + 1
            page_annots = page.annots()
            for annot in page_annots:
                if annot.type[1] == "Highlight":
                    annot_color = annot.colors["stroke"]
                    hex_color = rgb_to_hex(annot_color)
                    annot_rgb_fixed = tuple(int(color * 255) for color in annot_color)
                    if color != "":
                        dbs = get_db()
                        theme_data = (
                            dbs.query(Theme).filter(Theme.color == hex_color).first()
                        )
                        theme_color = theme_data.color
                        theme_term = theme_data.term
                        theme_hex_color = theme_data.hex
                        color_matched = is_color_close(annot_color, theme_hex_color)
                        if color_matched:
                            text = get_pdf_text(page, annot.rect)
                            if any(d["text"] == text for d in return_list):
                                continue
                            else:
                                total_count += 1
                                count += 1
                                item = {
                                    "ID": id,
                                    "count": count,
                                    "searched_color": color,
                                    "page_number": page_number,
                                    "annot_rgb_color": annot_rgb_fixed,
                                    "annot_hex_color": hex_color,
                                    "theme_hex_color": theme_hex_color,
                                    "theme_term": theme_term,
                                    "theme_color": theme_color,
                                    "text": text,
                                }
                                return_list.append(item)
                    else:
                        text = get_pdf_text(page, annot.rect)
                        if any(d["text"] == text for d in return_list):
                            continue
                        else:
                            total_count += 1
                            count += 1
                            item = {
                                "ID": id,
                                "count": count,
                                "page_number": page_number,
                                "annot_rgb_color": annot_rgb_fixed,
                                "annot_hex_color": hex_color,
                                "text": text,
                            }
                            return_list.append(item)

        if db:
            dbs = get_db()
            theme_id = dbs.query(Theme).filter(Theme.color == hex_color).first().id
            for i in return_list:
                dbs.add(
                    Annotation(
                        studies_id=id,
                        theme_id=theme_id,
                        count=i["count"],
                        page_number=i["page_number"],
                        annot_rgb_color=i["annot_rgb_color"],
                        annot_hex_color=i["annot_hex_color"],
                        text=i["text"],
                    )
                )
            dbs.commit()
            dbs.close()

    return total_count, return_list


##
## Extract Distribution
##


def extract_dist(pdf_path: str, term: str, id: str, db=False):
    """Extracts the distribution of the term from the pdf file and updates the distribution table in the database

    Args:
        pdf_path (path): The full path to the pdf file
        term (str): The term to search for in the pdf file
        id (int): Internal Study ID from the database e.g. Covidence number
        db (bool, optional): Save to the database? - Defaults to False.

    Returns:
        total_count (int), return_list (list): Total count of the term found in the pdf file, and the list of the distribution of the term
    """
    total_count = 0
    return_list = []

    with fitz.open(pdf_path) as doc:
        for page in doc:
            count = 0
            res = page.search_for(term)
            if res != []:
                for rect in res:
                    text = get_pdf_text(page, rect)
                    if any(d["text"] == text for d in return_list):
                        continue
                    else:
                        page_number = page.number + 1
                        # get citation from the database for id
                        dbs = get_db()
                        citation = (
                            dbs.query(Study)
                            .filter(Study.Covidence == id)
                            .first()
                            .citation
                        )
                        text = f" {text} {citation} page {page_number}."
                        total_count += 1
                        count += 1
                        item: dict = {
                            "studies_id": id,
                            "count": count,
                            "page_number": page_number,
                            "term": term,
                            "text": text,
                        }
                        return_list.append(item)

    if db:
        dbs = get_db()
        dbs.begin()
        dbs.query(Study).filter(Study.Covidence == id).update(
            {Study.total_distribution: total_count}
        )
        study_id = dbs.query(Study).filter(Study.Covidence == id).first().id
        theme_id = dbs.query(Theme).filter(Theme.term == term).first().id

        for i in return_list:
            # check if the text already exists in the distribution table
            if dbs.query(Distribution).filter(Distribution.text == i["text"]).first():
                continue
            else:
                dbs.add(
                    Distribution(
                        studies_id=study_id,
                        theme_id=theme_id,
                        count=i["count"],
                        page_number=i["page_number"],
                        term=i["term"],
                        text=i["text"],
                    )
                )
        dbs.commit()
        dbs.close()

    return total_count, return_list


def extract_total_dist_sheet_sync():
    """Extracts the total distribution from the database and updates the Distribution Worksheet in the Google Sheet

    Returns:
        bool: True if updated successfully, False if not
    """
    gs = get_conf("gs_url")
    ws = get_worksheet_by_name(gs, get_conf("default_studies"))
    headers_row_values = get_worksheet_headers_row_values(ws)
    id_col_values = get_worksheet_id_col_index_values(
        ws, get_conf("default_id")
    )
    updating_col_index_header = get_worksheet_updating_col_index_header(
        headers_row_values, "Distribution"
    )
    dbs = get_db()
    rows = dbs.query(Study).all()
    for row in rows:
        # update distribution column in google sheet from total_distribution column in database
        id = row.Covidence
        td = row.total_distribution  # db result
        try:
            # update sheet cell with td value where covidence number matches id
            res = update_sheet_cell(
                ws, id_col_values, id, updating_col_index_header, td
            )
            time.sleep(3)
        except:
            logger().warning(
                f"Google API rate limit reached, or other API error, please try again later!"
            )
            continue
    return res


def extract_dist_ws_sheet_sync(id: str):
    """Extracts the distribution from the database and updates the Distribution Worksheet in the Google Sheet

    Returns:
        bool: True if updated successfully, False if not
    """

    new_ws_name = f"Distribution-{get_random_string()}"
    sheet = get_spreadsheet_by_url(get_conf("gs_url"))
    new_ws = create_new_worksheet(sheet, new_ws_name)
    dbs = get_db()
    if id == "":
        dist_rows = dbs.query(Distribution).all()
    else:
        study_id = dbs.query(Study).filter(Study.Covidence == id).first()
        dist_rows = (
            dbs.query(Distribution)
            .filter(Distribution.studies_id == study_id.id)
            .all()
        )
    # dist_rows = dbs.query(Distribution).all()
    dist_rows_df = pd.DataFrame(
        [
            {
                "covidence_number": dbs.query(Study)
                .filter(Study.id == dist_row.studies_id)
                .first()
                .Covidence,
                "page_number": str(dist_row.page_number),
                "text": dist_row.text,
                "term": dist_row.term,
                "count": dist_row.count,
            }
            for dist_row in dist_rows
        ]
    )
    # add occurrence column, group by covidence_number and page_number
    dist_rows_df["occurrence"] = dist_rows_df.groupby(
        ["covidence_number", "page_number"]
    )["text"].transform("size")
    dist_rows_df["total_occurrence"] = dist_rows_df.groupby(["covidence_number"])[
        "text"
    ].transform("size")
    # convert covidence_number and page_number to int
    dist_rows_df["covidence_number"] = dist_rows_df["covidence_number"].astype(int)
    dist_rows_df["page_number"] = dist_rows_df["page_number"].astype(int)
    # sort by covidence_number and page_number
    dist_rows_df = dist_rows_df.sort_values(
        by=["covidence_number", "page_number"], ascending=True
    )
    dist_rows_df = dist_rows_df.reset_index(drop=True)
    # build a list for each id where text has same page number, add text to list
    dist_id_list = []
    for id in dist_rows_df["covidence_number"].unique():
        dist_id_list.append(
            dist_rows_df.loc[dist_rows_df["covidence_number"] == id].values.tolist()
        )
    # append dist_values to new_ws
    dist_values = dist_rows_df.values.tolist()
    for dist_value in dist_values:
        # covidence_number
        id = dist_value[0]
        # Searched term
        term = dist_value[3]
        # Total number of blocks the searched term found in the page
        total_occurence = dist_value[6]
        # Number of blocks the searched term found in the page
        occurernce = dist_value[5]
        # Page number
        page_number = dist_value[1]
        # total number of searched term found in the block of text
        count = dist_value[4]
        # block of text
        text = dist_value[2]
        try:
            res = new_ws.append_row(
                [
                    id,
                    term,
                    total_occurence,
                    occurernce,
                    page_number,
                    count,
                    text,
                ],
                include_values_in_response=True,
            )
            logger().info(f"Added to worksheet")
            # get the url for the first cell of the new appended row
            # cell = new_ws.find(id)
            # print(res)
            # cell =  res
            # cell_url = cell.url
            # res = total_dist_linker(id, cell_url)
            time.sleep(2)
        except:
            logger().warning(
                f"Google API rate limit reached, or other API error, please try again later!, res: {res}"
            )
            continue
    return True


# def total_dist_linker(id, dist_sheet_link):
#     # get the studies worksheet from google sheets
#     gs = get_conf("gs_url")
#     ws = get_worksheet_by_name(gs, get_conf("default_studies"))
#     # get the total distribution cell from the studies worksheet for the row that Covidence number matches id
#     id_col_values = get_worksheet_id_col_index_values(
#         ws, get_conf("default_id")
#     )
#     updating_col_index_header = get_worksheet_updating_col_index_header(
#         get_worksheet_headers_row_values(ws), "Distribution"
#     )
#     print(id_col_values)
#     print(updating_col_index_header)

#     # update the total distribution cell for the row that Covidence number matches id with the dist_sheet_link hyperlink
#     res = update_sheet_cell(
#         ws, id_col_values, id, updating_col_index_header, dist_sheet_link
#     )

#     return res



import typer
import os
import json
import sqlite3 as sql
import pyperclip

from rich import print
from typing_extensions import Annotated

# from scholarly import scholarly
from slh_sh.utils.file import get_conf

app = typer.Typer()


@app.command()
def info(
    id: Annotated[str, typer.Argument(help="ID, e.g. Covidence number")] = "",
    table: Annotated[str, typer.Option(help="Table name")] = get_conf("default_studies"),
    idcol: Annotated[str, typer.Option(help="ID column name")] = get_conf("default_id"),
    copy: Annotated[bool, typer.Option(help="Copy to clipboard")] = False,
):
    """Get info about a study and its Citation and Bibliography from a database table by ID."""

    if id == "":
        for file_name in os.listdir(get_conf("pdf_path")):
            if not file_name.startswith("."):
                print(file_name)
        print(
            f"""

There are {len(os.listdir(get_conf("pdf_path")))} PDFs in {get_conf("pdf_path")}

"""
        )
    else:
        conn = sql.connect(get_conf("sqlite_db"))
        curr = conn.cursor()

        curr.execute(f"SELECT * FROM {table} WHERE {idcol} = {id}")
        db_res = curr.fetchone()
        if db_res == None:
            print(f"Study with the ID: {id} not found in database table: {table}!")
        else:
            result = {}

            for i, col in enumerate(db_res):
                result[curr.description[i][0]] = col

            curr.execute(
                f"SELECT citation, bibliography FROM studies WHERE {get_conf('default_id')} = {id}"
            )
            db_citbib = curr.fetchone()

            conn.close()

            result["citation"] = db_citbib[0]
            result["bibliography"] = db_citbib[1]

            json_data = json.dumps(result, indent=4)
            print(json_data)

            if copy:
                text_output = ""
                for key, value in result.items():
                    if isinstance(value, str):
                        text_output += f"{key}: {value}\n"
                pyperclip.copy(text_output)

                print(
                    """

Copied to clipboard!

"""
                )



import sqlite3 as sql
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import Mapped, mapped_column, declared_attr
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import func
from datetime import datetime
from sqlalchemy import DateTime


sqlite_db_url = "sqlite:///slh.db"

# config_data = load_config()
Base = declarative_base()


class BaseModel(Base):
    __abstract__ = True

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), default=func.now(), nullable=False
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), default=func.now(), nullable=False
    )
    __name__: str

    # Generate __tablename__ automatically
    # @declared_attr
    # def __tablename__(cls) -> str:
    #     return cls.__name__.lower()


def get_db():
    """Gets the database session

    Returns:
        session (any): the database session
    """
    # create sqlalchemy engine and session
    engine = create_engine(sqlite_db_url)
    # db = engine.connect()
    Session = sessionmaker(bind=engine)
    session = Session()

    return session


# def get_db_cursor():  # DEPRECATED
#     """Gets the database cursor and connection

#     Returns:
#         conn (any), curr (any): the database connection and cursor
#     """
#     conn: sql.connect = sql.connect(sqlite_db_url)
#     curr: sql.Cursor = conn.cursor()
#     return conn, curr


# def table_exists(table_name: str) -> bool:  # DEPRECATED
#     curr = get_db_cursor()
#     curr.execute(
#         f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'"
#     )
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         return False
#     else:
#         return True


# def column_exists(table_name: str, column_name: str) -> bool:  # DEPRECATED
#     curr = get_db_cursor()
#     curr.execute(
#         f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'"
#     )
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         return False
#     else:
#         curr.execute(f"PRAGMA table_info({table_name})")
#         db_res: str = curr.fetchall()
#         for col in db_res:
#             if col[1] == column_name:
#                 return True
#         return False


# def create_db():  # DEPRECATED
#     """Creates the database and tables if they don't exist"""
#     conn, curr = get_db_cursor()
#     # check if Themes table exists
#     curr.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='themes'")
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         print(f"Themes table not found in database!")
#         # create Themes table
#         curr.execute(
#             """CREATE TABLE themes
#             (id INTEGER PRIMARY KEY AUTOINCREMENT,
#             studiesID INTEGER,
#             color TEXT NOT NULL,
#             hex TEXT NOT NULL,
#             term TEXT NOT NULL,
#             totalCount INTEGER,
#             FOREIGN KEY(studiesID) REFERENCES studies(id));"""
#         )
#         conn.commit()
#         print(f"Themes table created in database!")

#     # check if Searches table exists
#     curr.execute(
#         "SELECT name FROM sqlite_master WHERE type='table' AND name='searches'"
#     )
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         print(f"Searches table not found in database!")
#         # create Searches table
#         curr.execute(
#             """CREATE TABLE searches
#             (id INTEGER PRIMARY KEY AUTOINCREMENT,
#             name TEXT NOT NULL,
#             description TEXT NOT NULL);"""
#         )
#         conn.commit()
#         print(f"Searches table created in database!")

#     # check if Sources table exists
#     curr.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='sources'")
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         print(f"Sources table not found in database!")
#         # create Sources table
#         curr.execute(
#             """CREATE TABLE sources
#             (id INTEGER PRIMARY KEY AUTOINCREMENT,
#             name TEXT NOT NULL,
#             description TEXT NOT NULL);"""
#         )
#         conn.commit()
#         print(f"Sources table created in database!")

#     # annotations table: studiesID, themeID, annotation, pageNumber, pageImage
#     # check if Annotations table exists
#     curr.execute(
#         "SELECT name FROM sqlite_master WHERE type='table' AND name='annotations'"
#     )
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         print(f"Annotations table not found in database!")
#         # create Annotations table
#         curr.execute(
#             """CREATE TABLE annotations
#             (id INTEGER PRIMARY KEY AUTOINCREMENT,
#             studiesID INTEGER NOT NULL,
#             themeID INTEGER NOT NULL,
#             annotation TEXT NOT NULL,
#             pageNumber INTEGER NOT NULL,
#             text TEXT NOT NULL,
#             FOREIGN KEY(studiesID) REFERENCES studies(id),
#             FOREIGN KEY(themeID) REFERENCES themes(id));"""
#         )
#         conn.commit()
#         print(f"Annotations table created in database!")

#     # distribution table: id, studiesID, term, count, description (majority amount of the term under study headings), automatically calculated from themes table
#     # check if Distribution table exists
#     curr.execute(
#         "SELECT name FROM sqlite_master WHERE type='table' AND name='distribution'"
#     )
#     db_res: str = curr.fetchone()
#     if db_res == None:
#         print(f"Distribution table not found in database!")
#         # create Distribution table
#         curr.execute(
#             """CREATE TABLE distribution
#             (id INTEGER PRIMARY KEY AUTOINCREMENT,
#             studiesID INTEGER NOT NULL,
#             themeID INTEGER,
#             term TEXT NOT NULL,
#             pageNumber TEXT NOT NULL,
#             count INTEGER NOT NULL,
#             description TEXT,
#             FOREIGN KEY(themeID) REFERENCES themes(id),
#             FOREIGN KEY(studiesID) REFERENCES studies(id));"""
#         )
#         conn.commit()
#         print(f"Distribution table created in database!")



import time
import gspread

from oauth2client.service_account import ServiceAccountCredentials

from slh_sh.utils.file import get_conf, get_random_string
from slh_sh.utils.log import logger, get_now
from slh_sh.utils.db import get_db
from slh_sh.data.models import (
    Study,
)


def gs_auth():
    """Authenticate to Google Sheet API

    Returns:
        Client: Google Sheet Client
    """
    try:
        # Authorization
        scope = [
            "https://www.googleapis.com/auth/spreadsheets",
            "https://www.googleapis.com/auth/drive",
        ]
        credentials = ServiceAccountCredentials.from_json_keyfile_name(
            get_conf("google_credentials"), scope
        )
        gc = gspread.authorize(credentials)
        return gc
    except gspread.SpreadsheetNotFound:
        logger().warning(f"Spreadsheet authentication failed!")
        return "Spreadsheet authentication failed!"


def get_spreadsheet_by_url(gs):
    """Get Google Sheet by URL

    Args:
        gs (Client): Google Sheet Client

    Returns:
        Spreadsheet: Google Sheet
    """
    try:
        auth = gs_auth()
        spreadSheet = auth.open_by_url(gs)
        return spreadSheet
    except gspread.SpreadsheetNotFound:
        logger().warning(f"Spreadsheet {gs} not found!")
        return "Spreadsheet {gs} not found!"


def get_worksheet_by_name(gs, sheet):
    """Get Google Sheet's WordSheet by name

    Args:
        gs (Client): Google Sheet Client
        sheet (Worksheet): Google Sheet's Worksheet

    Returns:
        Worksheet (ws) or False: Google Sheet's Worksheet
    """
    try:
        spreadSheet = get_spreadsheet_by_url(gs)
        ws = spreadSheet.worksheet(sheet)
        return ws
    except gspread.WorksheetNotFound:
        logger().warning(f"Worksheet {sheet} not found in Google Sheet!")
        return "Worksheet {sheet} not found in Google Sheet!"


def get_worksheet_headers_row_values(ws):
    """Get Google Sheet's Worksheet's header row values

    Args:
        ws (Worksheet): Google Sheet's Worksheet

    Returns:
        Any: Google Sheet's Worksheet's header row values
    """
    header_row: int = int(get_conf("gs_header_row_number"))
    row_values = ws.row_values(header_row)

    return row_values


def get_worksheet_updating_col_index_header(header_row_values, sheetcol):
    """Get Google Sheet's Worksheet's updating column index header

    Args:
        header_row_values (list): Google Sheet's Worksheet's header row values
        sheetcol (str): Google Sheet's Worksheet's column name

    Returns:
        Any: Google Sheet's Worksheet's updating column index header
    """
    header_row_values = header_row_values.index(sheetcol)

    return header_row_values


def get_worksheet_id_col_index_values(ws, idcol):
    """Get Google Sheet's Worksheet's id column index values

    Args:
        ws (Worksheet): Google Sheet's Worksheet
        idcol (str): Google Sheet's Worksheet's id column name

    Returns:
        Any: Google Sheet's Worksheet's id column index values
    """
    header_row_values = get_worksheet_headers_row_values(ws)
    # get all values of id_col_index from header_row to the last row
    id_col_index = header_row_values.index(idcol)
    id_col_values = ws.col_values(id_col_index + 1)

    return id_col_values


def create_new_worksheet(sheet, name):
    """Create new Google Sheet's Worksheet

    Args:
        sheet (Worksheet): Google Sheet's Worksheet
        rows (str): Google Sheet's Worksheet's rows
        cols (str): Google Sheet's Worksheet's columns

    Returns:
        Worksheet: New Google Sheet's Worksheet
    """
    new_ws = sheet.add_worksheet(title=f"{name}", rows="1000", cols="200")
    print(f"New Worksheet created: {name}")

    return new_ws


def update_sheet_cell(
    ws, id_col_values, id_col_value, updating_col_index_header, new_value
):
    """Update Google Sheet's Worksheet's cell

    Args:
        ws (Worksheet): Google Sheet's Worksheet
        id_col_values (Any): Google Sheet's Worksheet's id column values
        id_col_value (Any): Google Sheet's Worksheet's id column value
        updating_col_index_header (Any): Google Sheet's Worksheet's updating column index header
        new_value (Any): Value to be updated

    Returns:
        bool: True if updated successfully, False if not
    """
    try:
        ws.update_cell(
            id_col_values.index(id_col_value) + 1,
            updating_col_index_header + 1,
            new_value,
        )
        print(f"Updated {id_col_value} with '{new_value}'")
        logger().info(f"{get_now()} Updated {id_col_value} with '{new_value}'")
        return True
    except gspread.exceptions.APIError:
        logger().warning(
            f"Google API rate limit reached, or other API error, please try again later!"
        )
        return False


def sync_studies_sheet(gs):
    """Sync studies table of sqlite database to Google Sheet

    Args:
        gs (str): Google Sheet URL

    Returns:
        str: New Worksheet title
    """
    sheet = get_spreadsheet_by_url(gs)
    new_ws_name = f"slh-{get_random_string()}"
    new_ws = sheet.add_worksheet(title=f"{new_ws_name}", rows="1000", cols="200")
    print(f"New Worksheet created: {new_ws_name}")

    dbs = get_db()
    rows = dbs.query(Study).all()
    col_names = Study.__table__.columns.keys()
    try:
        new_ws.append_row(col_names)
    except gspread.exceptions.APIError:
        logger().warning(
            f"Google API rate limit reached, or other API error, please try again later!"
        )
        time.sleep(3)

    # add all data from studies table of sqlite dababase to the new Worksheet
    for row in rows:
        row = list(row)
        try:
            new_ws.append_row(row)
            print(f":white_check_mark: {row[0]}")
        except:
            print(f":x: {row[0]}")
            logger().warning(
                f"Google API rate limit reached, or other API error, please try again later!"
            )
            continue

        # sleep for 3 seconds to avoid Google API rate limit
        time.sleep(3)

    return new_ws.title


def sync_studies_column_sheet(ws, idcol, sheetcol, apply=False):
    """Sync studies table of sqlite database to Google Sheet

    Args:
        ws (Worksheet): Google Sheet's Worksheet
        idcol (str): Google Sheet's Worksheet's id column name
        sheetcol (str): Google Sheet's Worksheet's column name
        apply (bool, optional): Defaults to False.

    Returns:
        bool: True if updated successfully, False if not
    """
    # select all Keywords from database where Covidence is id_col_values
    updating_col_index_header = get_worksheet_updating_col_index_header(
        get_worksheet_headers_row_values(ws), sheetcol
    )
    id_col_values = get_worksheet_id_col_index_values(ws, idcol)
    for id_col_value in id_col_values[3:]:
        dbs = get_db()
        # select [idcol = covidence] from database where idcol is id_col_value
        db_res = dbs.query(Study).filter_by(idcol=id_col_value).first()
        # check if SELECT found any value
        if db_res == None:
            print(f"Study {id_col_value} not found in database!")
            continue
        elif db_res[0] != None:
            if apply:
                try:
                    update_sheet_cell(
                        ws,
                        id_col_values,
                        id_col_value,
                        updating_col_index_header,
                        db_res[0],
                    )
                    time.sleep(3)
                except gspread.exceptions.APIError:
                    logger().warning(
                        f"Google API rate limit reached, or other API error, please try again later!"
                    )
                    return False
                # sleep for 3 seconds to avoid Google API rate limit
            else:
                print(f"Would update {id_col_value} with '{db_res[0]}'")
        else:
            print(f"Empty {sheetcol} for {id_col_value}, skipping...")

    return True



import typer
import sqlite3 as sql
import csv as csvimport

from typing_extensions import Annotated
from rich import print

from slh_sh.utils.file import get_conf

app = typer.Typer()


@app.command("csv")  # TODO: REFACTOR
# TODO: Adding new CSV file creates new DB, asks for DB name, and adds them to Config.yaml
def csv(
    csv: Annotated[
        str,
        typer.Argument(
            help=f"""
        Imports a CSV file into a SQLite database.

        1. Export the CSV file  from Covidence via Export > REFERENCES > Options [Full text review] > Format [CSV]
2. Rename to {get_conf("csv_export")}
3. Place it in the project directory.
    """
        ),
    ] = get_conf("csv_export"),
):
    input(
        f"""
        Press Enter to import:

        CSV File: {csv}
        Sqlite Database Name: {get_conf("sqlite_db")}

        Press Ctrl+C to cancel.
        """
    )
    print(
        f"""
        Importing {csv} to {get_conf('sqlite_db')}...
        """
    )

    with open(get_conf("csv_export"), "r") as f:
        reader = csvimport.reader(f)
        # Read the first line of the file to get the CSV headers
        headers = next(reader)

        headers = [header.replace(" ", "_").replace("_#", "") for header in headers]

        conn = sql.connect(get_conf("sqlite_db"))
        curr = conn.cursor()

        # check if default_id is not empty in config.yaml
        if get_conf("default_id") == "":
            print(
                "default_id is empty in config.yaml incremental number will be used as id."
            )
            # Create a new table in the SQLite database with the same columns as the CSV file headers and add an id column
            curr.execute(
                f"""CREATE TABLE IF NOT EXISTS studies (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                {", ".join(headers)}
            )"""
            )
        else:
            # Create a new table in the SQLite database with the same columns as the CSV file headers
            curr.execute(
                f"""CREATE TABLE IF NOT EXISTS studies (
                {", ".join(headers)}
            )"""
            )

            # Iterate over the rows of the CSV file and insert each row into the table in the database
            rows = []
            for row in reader:
                rows.append(tuple(row))

            # check if the study is already in the database based on the covidence number as id if not insert it
            for row in rows:
                curr.execute(
                    f"SELECT * FROM studies WHERE ? = ?", (get_conf("default_id"), row[12],)
                )
                if len(curr.fetchall()) > 0:
                    print(
                        f":information_desk_person: {row[12]} is already in the database."
                    )
                else:
                    curr.execute(
                        # insert row into sqlite db
                        f"INSERT INTO studies ({', '.join(headers.lower())}) VALUES ({', '.join([f'?' for i in range(len(headers.lower()))])})",
                        row,
                    )
                    print(f":star: {row[12]} added to the database.")

    print(
        f"""
        :tada: {len(rows)} studies processed.

        """
    )

    f.close()
    conn.commit()
    conn.close()




import os
import yaml
import random
import string

from pathlib import Path


def get_conf(key: str) -> str:
    """Returns the value of the given key in the config file.

    Args:
        key (str): Key in the config file.

    Returns:
        str: Value of the given key.
    """
    config_path = Path.cwd() / "config.yaml"
    if config_path.exists():
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)
    else:
        print(
            """
            [bold green]Good day researcher :wave:[/bold green]

            The [red]config.yaml[/red] file does not exist in the current directory.

                - Run [yellow]slh self init[/yellow] to initialize a new SLR project :rocket:
                - [blue]
            """
        )

    return config[key]


def get_pdf_dir() -> str:
    """Returns the path to the PDF directory.

    Returns:
        str: Path to the PDF directory.
    """
    return Path.cwd() / get_conf("pdf_path")


def get_file_path(id):
    """Return the file path for the given covidence number.

    Args:
        id: ID column of the study specified in the config file and exists in the file name.
    """
    pdf_dir = get_pdf_dir()
    pdf_path = None
    pdf_id = None

    for file_name in os.listdir(pdf_dir):
        pdf_id: str = file_name.split("_")[0].replace("#", "")
        if pdf_id == str(id):
            pdf_path = os.path.join(pdf_dir, file_name)

    return pdf_path


def file_name_generator(id: str, authors: str, year: str) -> list[str]:
    """Generates a file name for the given covidence number, authors and year.

    Args:
        id (str): ID column of the study specified in the config file and exists in the file name e.g. Covidence Number.
        authors (str): Authors column
        year (str): Publication year column

    Returns:
        list[str]: Retuns a list of file names.
    """
    if ";" not in authors:
        last_name: str = authors.split(",")[0].strip()
        name: str = f"{id}_{last_name}_{year}"
        return name
    elif authors.count(";") == 1:
        last_name_1: str = authors.split(";")[0].split(",")[0].strip()
        last_name_2: str = authors.split(";")[1].split(",")[0].strip()
        name: str = f"{id}_{last_name_1}_{last_name_2}_{year}"
        return name
    elif authors.count(";") > 1:
        last_name: str = authors.split(";")[0].split(",")[0].strip()
        name = f"{id}_{last_name}_et_al_{year}"
        return name
    else:
        return None


def get_random_string():
    """Generates a random string.

    Returns:
        str: Random lowercase 4 letters string.
    """
    random_string = "".join(random.choices(string.ascii_lowercase, k=4))
    return random_string



import os
import logging
import datetime
from logging.handlers import RotatingFileHandler


def logger():
    """Create a logger object and return it."""

    # create ./logs directory if it doesn't exist
    if not os.path.exists("./logs"):
        os.makedirs("./logs")

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s %(levelname)s: %(message)s",
    )
    logger = logging.getLogger(__name__)
    # remove old handlers
    logger.handlers = []

    file_handler = logging.FileHandler("./logs/app.log")  # TODO: make this configurable

    logger.addHandler(file_handler)

    # make logger only logs to file and not console
    logger.propagate = False

    # rotate logs every 10MB
    file_handler = RotatingFileHandler(
        "./logs/app.log", maxBytes=10 * 1024 * 1024, backupCount=5
    )
    return logger


def get_now():
    """Get current date and time"""
    now = datetime.datetime.now()
    fnow = now.strftime("%Y-%m-%d %H:%M:%S")
    return fnow



import typer
import sqlite3 as sql
import pyperclip
import json

from rich import print
from typing_extensions import Annotated

from slh_sh.utils.file import get_conf

app = typer.Typer()


@app.command()
def query(
    terms: Annotated[
        list[str],
        typer.Argument(
            help="""Multipart with one or more arguments,
first argument one or more Author or ID separated by colon, second ThemeName, third SubThemeName
e.g. slh-sh query themes John,120,192 Stage_1 Derogation or a direct SQL query with -s/--sqlquery"""
        ),
    ],
    studytable: Annotated[
        str, typer.Option("-st", "--studytable", help="Study table name")
    ] = get_conf("default_studies"),
    idname: Annotated[
        str, typer.Option("-id", "--idname", help="ID column name")
    ] = get_conf("default_id"),
    sqlquery: Annotated[
        bool, typer.Option("-s", "--sqlquery", help="SQL query")
    ] = False,
    copy: Annotated[
        bool, typer.Option("-c", "--copy", help="Copy to clipboard")
    ] = False,
):
    """Get themes about a study from a database table by ID."""

    if sqlquery != False:
        conn = sql.connect(get_conf("sqlite_db"))
        curr = conn.cursor()
        print(terms[0])
        curr.execute(terms[0])
        db_res = curr.fetchall()
        if db_res == None:
            print(
                f"""
Nothing Found!
sqlquery: {sqlquery}
result: {db_res}
"""
            )
        else:
            db_res = [
                dict(zip([col[0] for col in curr.description], row)) for row in db_res
            ]
            json_data = json.dumps(db_res, indent=4)
            print(json_data)
            if copy:
                pyperclip.copy(json_data)
                print("Copied to clipboard!")
        conn.close()
        exit()

    authors = terms[0].split(",")
    theme = ""
    subtheme = ""
    if len(terms) == 1:
        input = terms[0].lower()
        if input and " " in input:
            print(
                """
SQL query detected!
to query sql database directly use -s/--sqlquery option, e.g.:
slh-sh query -s "SELECT * FROM studies WHERE authors LIKE '%John%'"
"""
            )
            exit()
        else:
            pass
    if len(terms) == 2:
        theme = terms[1]
    if len(terms) == 3:
        theme = terms[1]
        subtheme = terms[2]
    if len(terms) > 3:
        print("Too many arguments!")
        exit()

    results = {}

    conn = sql.connect(get_conf("sqlite_db"))
    curr = conn.cursor()

    for author in authors:
        if author.isnumeric():
            curr.execute(f"SELECT * FROM {studytable} WHERE {idname} = {author}")
            db_res = curr.fetchone()
            if db_res == "None":
                print(
                    f"Study with the ID: {author} not found in database table: {studytable}!"
                )
            else:
                curr.execute(
                    f"SELECT citation, bibliography FROM studies WHERE {get_conf('default_id')} = {author}"
                )
                db_citbib = curr.fetchone()
                res = {
                    "ID": author,
                    "Authors": db_res[2],
                    "Title": db_res[1],
                    "Theme": theme,
                    "Subtheme": subtheme,
                    "Citation": db_citbib[0],
                    "Bibliography": db_citbib[1],
                }
                results[author] = res
        else:
            curr.execute(f"SELECT * FROM {studytable} WHERE authors LIKE '%{author}%'")
            db_res = curr.fetchone()
            if db_res == "None":
                print(
                    f"Study with the Author: {author} not found in database table: {studytable}!"
                )
            else:
                id_index = [
                    i for i, col in enumerate(curr.description) if col[0] == idname
                ][0]
                curr.execute(
                    f"SELECT citation, bibliography FROM studies WHERE {get_conf('default_id')} = {db_res[id_index]}"
                )
                db_citbib = curr.fetchone()
                res = {
                    "ID": db_res[id_index],
                    "Authors": db_res[2],
                    "Title": db_res[1],
                    "Theme": theme,
                    "Subtheme": subtheme,
                    "Citation": db_citbib[0],
                    "Bibliography": db_citbib[1],
                }
                results[author] = res
    if theme != "":
        for key, value in results.items():
            if subtheme == "":
                curr.execute(f"SELECT * FROM {theme} WHERE {idname} = {value['ID']}")
                db_res = curr.fetchone()
                if db_res == None:
                    print(
                        f"Theme for Study with the ID: {value['ID']} not found in database table: {theme}!"
                    )
                else:
                    # add column names to result
                    for i, col in enumerate(db_res):
                        value[curr.description[i][0]] = col
            else:
                curr.execute(
                    f"SELECT {theme},{subtheme} FROM {theme} WHERE {idname} = {value['ID']}"
                )
                db_res = curr.fetchone()
                if db_res == None:
                    print(
                        f"Subtheme for Study with the ID: {value['ID']} not found in database table: {theme}!"
                    )
                else:
                    value["Theme_Text"] = db_res[0]
                    value["Subtheme_Text"] = db_res[1]
    else:
        print("No theme specified!")

    conn.close()
    print(results)

    if copy:
        text = ""
        if subtheme != "":
            text = value["Subtheme_Text"] + " " + value["Citation"]
        elif theme != "":
            text = value[theme] + " " + value["Citation"]
        else:
            # add citation from results
            for key, value in results.items():
                text += value["Citation"] + "\n"
        pyperclip.copy(text)
        print(
            """

The following copied to clipboard!
Note: If query was for multiple studies, only the last result was copied.

"""
        )
        print(text + "\n")



from pydantic import BaseModel


class StudySchema(BaseModel):
    id: int
    title: str
    authors: list
    abstract: str
    published_year: int
    published_month: int
    journal: str
    volume: str
    issue: str
    pages: str
    accession_number: str
    doi: str
    ref: str
    Covidence: int
    study: str
    notes: str
    tags: str
    filename: str
    keywords: str
    citation: str
    bibliography: str
    full_text: str
    total_annotations: int
    total_distribution: int


class ThemeSchema(BaseModel):
    id: int
    name: str
    description: str
    color: str
    studies: list
    annotations: list
    distribution: list


class AnnotationSchema(BaseModel):
    id: int
    studies_id: int
    theme_id: int
    count: int
    page_number: int
    annot_rgb_color: str
    annot_hex_color: str
    annotation: str
    text: str


class DistributionSchema(BaseModel):
    id: int
    studies_id: int
    theme_id: int
    count: int
    page_number: int
    term: str
    text: str



import typer
import sys
import sqlite3 as sql

# import sqlalchemy as sa

# from sqlalchemy.inspection import inspect

from rich import print
from typing_extensions import Annotated

from slh_sh.utils.log import logger
from slh_sh.utils.file import get_conf
from slh_sh.modules.sync import (
    get_worksheet_by_name,
    update_sheet_cell,
    sync_studies_sheet,
    sync_studies_column_sheet,
    get_worksheet_id_col_index_values,
    get_worksheet_updating_col_index_header,
    get_worksheet_headers_row_values,
)
from slh_sh.utils.db import get_db
from slh_sh.data.models import (
    Study,
)

app = typer.Typer()


@app.command()
def update(
    sheetcol: Annotated[str, typer.Option(help="Column name in Google Sheet")] = "",
    # TODO: handle range when passed by user as id, Use a number, a range e.g. 1-20 or multiple numbers e.g 1,30,2
    id: Annotated[str, typer.Option(help="ID, e.g. Covidence number")] = "",
    sheet: Annotated[str, typer.Option(help="Name of the Sheet")] = get_conf(
        "default_studies"
    ),
    gs: Annotated[str, typer.Option(help="Google Sheet URL")] = get_conf("gs_url"),
    idcol: Annotated[
        str,
        typer.Option(help="Column name that considered as ID e.g 'Covidence Number'"),
    ] = get_conf("default_id"),
    apply: Annotated[bool, typer.Option(help="Apply changes to Google Sheet")] = False,
    allcol: Annotated[bool, typer.Option(help="Update all data in column")] = False,
    alltable: Annotated[
        str, typer.Option(help="Add all data of a Table to a new Worksheet")
    ] = "",
):
    """
    Update Google Sheet with data from Database.

    --id <default_id> --gs <sheet_name> --col <column_name>
    """
    if alltable == "" and apply:
        ws = get_worksheet_by_name(gs, sheet)
    if id != "" and sheetcol != "" and "," not in id and "-" not in id:
        # get Google Sheet's Worksheet's id column index values
        headers_row_values = get_worksheet_headers_row_values(ws)
        id_col_values = get_worksheet_id_col_index_values(ws, idcol)
        updating_col_index_header = get_worksheet_updating_col_index_header(
            headers_row_values, sheetcol
        )
        # curr.execute(f"SELECT {sheetcol} FROM studies WHERE Covidence = '{id}'")
        # db_res: str = curr.fetchone()
        dbs = get_db()
        # select [idcol = covidence] from database where idcol is id_col_value
        db_res = dbs.query(Study).filter_by(idcol=id_col_values).first()
        if db_res == None:
            print(f"Study {id} not found in database!")
            sys.exit()
        elif db_res[0] != None:
            if apply:
                update_sheet_cell(
                    ws, id_col_values, id, updating_col_index_header, db_res[0]
                )
                print(
                    """
            :tada: Sync finished successfully from database to Google Sheet:
            Column: {col}
            Sheet: {sheet}
            Study: {id}
            Google Sheet {gs}...
                    """
                )
            else:
                print(f"Would update {id} with '{db_res[0]}'")
        else:
            print(f"Empty {sheetcol} for {id}, skipping...")
    elif allcol and sheetcol != "":
        res = sync_studies_column_sheet(gs, idcol, sheetcol)
        print(
            f"""
            :tada: Sync finished successfully from database to Google Sheet:
            Column: {sheetcol}
            Sheet: {sheet}
            Google Sheet {gs}...
            """
        )
    elif alltable != "":
        if apply:
            print(f"Appending all data in the {alltable} to a new Worksheet...")
            print(gs)
            new_worksheet_title = sync_studies_sheet(gs)
            print(
                f"""
            :tada: Sync finished successfully from database to Google Sheet:
            New Worksheet: {new_worksheet_title}
            Table: {alltable}
            Google Sheet {gs}...
            """
            )
        else:
            print(
                f"""
                By adding --apply the following will happen:
                1. Would create a new Worksheet at Google Sheet from database...
                2. Would add all data from {alltable} table of sqlite dababase to the new Worksheet
                """
            )
    else:
        print(f"Invalid covidence id {id}, {sheetcol} or table name {alltable}!")
        # if id contains , or - then it's a range
        # if "," in id:
        #     print("Multiple detected!")
        #     id = id.split(",")
        #     print(id)
        #     # check if any value in id exists in id_col_values
        #     if apply:
        #         for id_col_value in id_col_values[3:]:
        #             if id in id_col_value:
        #                 print(id)
        #                 update_sheet_cell(
        #                     ws, id_col_values, id_col_value, updating_col_index_header, db_res[0])
        #                 # sleep for 3 seconds to avoid Google API rate limit
        #                 time.sleep(3)
        #             else:
        #                 print(f"Study {id_col_value} not found in database!")
        #                 continue
        #     sys.exit()
        # elif "-" in id:
        #     print("Range detected!")
        #     id = id.split("-")
        #     print(id)
        #     sys.exit()


@app.command()
def config():
    """
    Iterate over Themes, Searches, and Sources in config.yaml and insert into database.

    Creates the needed tables if not exists.
    """
    print(f"Syncing config.yaml to database: {get_conf('sqlite_db')}...")

    conn: sql.connect = sql.connect(get_conf("sqlite_db"))
    curr: sql.Cursor = conn.cursor()

    # iterate over Themes in config_data and insert into database
    for theme in get_conf("themes"):
        # check if theme exists in database
        curr.execute(f"SELECT id FROM themes WHERE color = '{theme}'")
        db_res: str = curr.fetchone()
        if db_res == None:
            # insert theme into database
            hex = get_conf("themes")[theme]["hex"]
            term = get_conf("themes")[theme]["term"]
            # TODO: add study (id) and totalCount columns
            curr.execute(
                f"INSERT INTO themes (color, hex, term) VALUES ('{theme}', '{hex}', '{term}');"
            )
            conn.commit()
            print(f"Theme {theme} inserted in database!")
        else:
            print(f"Theme {theme} already exists in database!")

    # iterate over Searches in config_data and insert into database
    for search in get_conf("searches"):
        # check if search exists in database
        curr.execute(f"SELECT id FROM searches WHERE name = '{search}'")
        db_res: str = curr.fetchone()
        if db_res == None:
            # insert search into database
            description = get_conf("searches")[search]
            curr.execute(
                f"INSERT INTO searches (name, description) VALUES ('{search}', '{description}');"
            )
            conn.commit()
            print(f"Search {search} inserted in database!")
        else:
            print(f"Search {search} already exists in database!")

    # iterate over Sources in config_data and insert into database
    for source in get_conf("sources"):
        # check if source exists in database
        curr.execute(f"SELECT id FROM sources WHERE name = '{source}'")
        db_res: str = curr.fetchone()
        if db_res == None:
            # insert source into database
            description = get_conf("sources")[source]
            curr.execute(
                f"INSERT INTO sources (name, description) VALUES ('{source}', '{description}');"
            )
            conn.commit()
            print(f"Source {source} inserted in database!")
        else:
            print(f"Source {source} already exists in database!")

    conn.close()

@app.command()
def fetch(
    sheet: Annotated[str, typer.Argument(help="Name of the Sheet e.g. 'Stage 1' or Stage_1")] = "",
    dbtable: Annotated[
        str,
        typer.Argument(
            help="Name of the Database Table, if empty the name will be used e.g Stage_One_Sheet"
        ),
    ] = "",
    gs: Annotated[str, typer.Option(help="Google Sheet URL")] = get_conf("gs_url"),
):
    """Fetch data from Google Sheet to a new database table."""

    print(f"Fetching Google Sheet to database...")

    ws = get_worksheet_by_name(gs, sheet)
    ws_data = ws.get_values()
    ws_headers = ws_data[2]
    ws_data = ws_data[3:]

    if dbtable == "":
        dbtable = sheet
        sheet = sheet.replace(" ", "_")
    elif dbtable != "":
        dbtable = dbtable.replace(" ", "_")

    ws_headers = [header.replace(" ", "_") for header in ws_headers]
    ws_headers = [header.replace("/", "_") for header in ws_headers]

    # replace empty string with None in ws_data
    ws_data = [["None" if cell == "" else cell for cell in row] for row in ws_data]

    conn = sql.connect(get_conf("sqlite_db"))
    curr = conn.cursor()

    curr.execute(
        f""" CREATE TABLE IF NOT EXISTS {dbtable} ({', '.join([f'{header} TEXT' for header in ws_headers])});"""
    )
    conn.commit()

    for row in ws_data:
        print(row)
        curr.execute(
            f"INSERT INTO {dbtable} ({', '.join(ws_headers)}) VALUES ({', '.join([f'?' for i in range(len(ws_headers))])})",
            row,
        )
        conn.commit()

    conn.close()

    print(f"Data from {dbtable} added to corresponding database table!")



import typer
import webbrowser
import sqlite3 as sql

from pathlib import Path
from rich import print
from typing_extensions import Annotated

from slh_sh.utils.file import get_file_path, get_conf

app = typer.Typer()


@app.command()
def gd(
    url: Annotated[str, typer.Argument(help="Google Drive URL")] = get_conf("gd_url")
):
    print(f"Opening Google Drive: {url}...")
    webbrowser.open(url)


@app.command()
def gs(
    url: Annotated[str, typer.Argument(help="Google Sheet URL")] = get_conf("gs_url")
):
    print(f"Opening Google Sheet: {url}...")
    webbrowser.open(url)


@app.command()
# https://www.foxit.com/pdf-reader/
def pdf(
    id: Annotated[str, typer.Argument(help="ID, e.g. Covidence number")] = "",
    # page: Annotated[str, typer.Option(help="Page number")] = "",
):
    pdf_path = get_file_path(id)
    print(pdf_path)
    if pdf_path is None:
        print(f"PDF not found for {id}")
        typer.Exit()
    else:
        typer.launch(pdf_path)
    ##
    ## Opens a PDF on certain page, Not Supported by PDF reader Apps or OS
    ##
    # if page != "":
    #     # open the pdf file with the page number with local pdf reader
    #     # open_pdf(id, page)
    #     print(pdf_path)
    #     # check if os is windows
    #     if os.name == "nt":
    #         subprocess.run(["acroread", "--goto", str(page), pdf_path])
    #     # check if os is mac
    #     # elif os.name == "darwin":
    #     #     subprocess.run(["open", "-a", "Preview", pdf_path])
    #     # check if os is linux
    #     # elif os.name == "posix":
    #     #     subprocess.run(["xdg-open", pdf_path])
    #     else:
    #         print("OS not supported")


@app.command()
# https://sqlitebrowser.org/dl/
def db(
    sql: Annotated[str, typer.Argument(help="Name of SQLite database")] = get_conf(
        "sqlite_db"
    )
):
    print(f"Opening SQLite database: {sql}...")

    sql_path = Path.cwd() / sql

    if sql_path is None:
        print(f"SQLite database not found")
        typer.Exit()
    webbrowser.open(sql_path.as_uri())


@app.command()
def doi(
    id: Annotated[
        str, typer.Argument(help="ID of the study e.g Covidence Number")
    ] = get_conf("default_id"),
):
    """Opens the DOI of a study by ID in browser."""

    conn = sql.connect(get_conf("sqlite_db"))
    curr = conn.cursor()
    curr.execute(f"SELECT doi FROM studies WHERE {get_conf('default_id')} = {id}")
    doi = curr.fetchone()
    conn.close()
    if doi == None:
        print(f"DOI with the ID: {id} not found in database table: studies!")
    else:
        doi = doi[0]
        print(f"Opening DOI: {doi}...")
        webbrowser.open(f"https://doi.org/{doi}")
